@@ -15,6 +15,7 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <cairo/cairo-pdf.h>
 #include <cairo/cairo-ps.h>
+#include <unistd.h> // TEMP: Needed for close() in get_temporary_filename()
 
 #include "book.h"
 
@@ -62,8 +63,7 @@ book_append_page (Book *book, gint width, gint height, gint dpi, Orientation ori
     Page *page;
 
     page = page_new ();
-    page_set_scan_area (page, width, height, dpi);
-    page_set_orientation (page, orientation);
+    page_setup (page, width, height, dpi, orientation);
 
     book->priv->pages = g_list_append (book->priv->pages, page);
 
@@ -76,8 +76,9 @@ book_append_page (Book *book, gint width, gint height, gint dpi, Orientation ori
 void
 book_delete_page (Book *book, Page *page)
 {
-    book->priv->pages = g_list_remove (book->priv->pages, page);
     g_signal_emit (book, signals[PAGE_REMOVED], 0, page);
+
+    book->priv->pages = g_list_remove (book->priv->pages, page);
     g_object_unref (page);
 }
 
@@ -98,33 +99,14 @@ book_get_page (Book *book, gint page_number)
 }
 
 
-static gboolean
-write_pixbuf_data (const gchar *buf, gsize count, GError **error, GFileOutputStream *stream)
-{
-    return g_output_stream_write_all (G_OUTPUT_STREAM (stream), buf, count, NULL, NULL, error);
-}
-
-
-static GFileOutputStream *
-open_file (const gchar *uri, GError **error)
-{
-    GFile *file;
-    GFileOutputStream *stream; 
-
-    file = g_file_new_for_uri (uri);
-    stream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, error);
-    g_object_unref (file);
-    return stream;
-}
-
-
-static gchar *
-make_indexed_uri (const gchar *uri, gint i)
+static GFile *
+make_indexed_file (const gchar *uri, gint i)
 {
     gchar *basename, *suffix, *indexed_uri;
+    GFile *file;
 
     if (i == 0)
-        return g_strdup (uri);
+        return g_file_new_for_uri (uri);
 
     basename = g_path_get_basename (uri);
     suffix = g_strrstr (basename, ".");
@@ -134,70 +116,32 @@ make_indexed_uri (const gchar *uri, gint i)
     else
         indexed_uri = g_strdup_printf ("%s-%d", uri, i);
     g_free (basename);
-    return indexed_uri;
+
+    file = g_file_new_for_uri (indexed_uri);
+    g_free (indexed_uri);
+
+    return file;
 }
 
 
-gboolean
-book_save_jpeg (Book *book, const gchar *uri, GError **error)
+static gboolean
+book_save_multi_file (Book *book, const gchar *type, GFile *file, GError **error)
 {
     GList *iter;
     gboolean result = TRUE;
     gint i;
+    gchar *uri;
 
-    for (iter = book->priv->pages, i = 0; iter; iter = iter->next, i++) {
+    uri = g_file_get_uri (file);
+    for (iter = book->priv->pages, i = 0; iter && result; iter = iter->next, i++) {
         Page *page = iter->data;
-        GFileOutputStream *stream;
-        GdkPixbuf *image;
-        gchar *indexed_uri;
+        GFile *file;
 
-        indexed_uri = make_indexed_uri (uri, i);
-        stream = open_file (indexed_uri, error);
-        g_free (indexed_uri);
-        if (!stream)
-            return FALSE;
-
-        image = page_get_cropped_image (page);
-        result = gdk_pixbuf_save_to_callback (image,
-					      (GdkPixbufSaveFunc) write_pixbuf_data, stream,
-					      "jpeg", error,
-					      "quality", "90",
-					      NULL);
-        g_object_unref (image);
-        g_object_unref (stream);
-    }
-   
-    return result;
-}
-
-
-gboolean
-book_save_png (Book *book, const gchar *uri, GError **error)
-{
-    GList *iter;
-    gboolean result = TRUE;
-    gint i;
-
-    for (iter = book->priv->pages, i = 0; iter; iter = iter->next, i++) {
-        Page *page = iter->data;
-        GFileOutputStream *stream;
-        GdkPixbuf *image;
-        gchar *indexed_uri;
-
-        indexed_uri = make_indexed_uri (uri, i);
-        stream = open_file (indexed_uri, error);
-        g_free (indexed_uri);
-        if (!stream)
-            return FALSE;
-
-        image = page_get_cropped_image (page);
-        result = gdk_pixbuf_save_to_callback (image,
-					      (GdkPixbufSaveFunc) write_pixbuf_data, stream,
-					      "png", error,
-					      NULL);
-        g_object_unref (image);
-        g_object_unref (stream);
+        file = make_indexed_file (uri, i);
+        result = page_save (page, type, file, error);
+        g_object_unref (file);
     }
+    g_free (uri);
    
     return result;
 }
@@ -236,14 +180,14 @@ write_cairo_data (GFileOutputStream *stream, unsigned char *data, unsigned int l
 }
 
 
-gboolean
-book_save_ps (Book *book, const gchar *uri, GError **error)
+static gboolean
+book_save_ps (Book *book, GFile *file, GError **error)
 {
     GFileOutputStream *stream;
     GList *iter;
     cairo_surface_t *surface;
 
-    stream = open_file (uri, error);
+    stream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, error);
     if (!stream)
         return FALSE;
 
@@ -274,14 +218,154 @@ book_save_ps (Book *book, const gchar *uri, GError **error)
 }
 
 
-gboolean
-book_save_pdf (Book *book, const gchar *uri, GError **error)
+// TEMP: Copied from simple-scan.c
+static GFile *
+get_temporary_file (const gchar *prefix, const gchar *extension)
+{
+    gint fd;
+    GFile *file;
+    gchar *filename, *path;
+    GError *error = NULL;
+
+    /* NOTE: I'm not sure if this is a 100% safe strategy to use g_file_open_tmp(), close and
+     * use the filename but it appears to work in practise */
+
+    filename = g_strdup_printf ("%s-XXXXXX.%s", prefix, extension);
+    fd = g_file_open_tmp (filename, &path, &error);
+    g_free (filename);
+    if (fd < 0) {
+        g_warning ("Error saving email attachment: %s", error->message);
+        g_clear_error (&error);
+        return NULL;
+    }
+    close (fd);
+  
+    file = g_file_new_for_path (path);
+    g_free (path);
+
+    return file;
+}
+
+
+static goffset
+get_file_size (GFile *file)
+{
+    GFileInfo *info;
+    goffset size = 0;
+  
+    info = g_file_query_info (file,
+                              G_FILE_ATTRIBUTE_STANDARD_SIZE,
+                              G_FILE_QUERY_INFO_NONE,
+                              NULL,
+                              NULL);
+    if (info) {
+        size = g_file_info_get_size (info);
+        g_object_unref (info);
+    }
+
+    return size;
+}
+
+
+static gboolean
+book_save_pdf_with_imagemagick (Book *book, GFile *file, GError **error)
+{
+    GList *iter;
+    GString *command_line;
+    gboolean result = TRUE;
+    gint exit_status = 0;
+    GFile *output_file = NULL;
+    GList *link, *temporary_files = NULL;
+
+    /* ImageMagick command to create a PDF */
+    command_line = g_string_new ("convert -adjoin");
+
+    /* Save each page to a file */
+    for (iter = book->priv->pages; iter && result; iter = iter->next) {
+        Page *page = iter->data;
+        GFile *jpeg_file, *tiff_file;
+        gchar *path;
+        gint jpeg_size, tiff_size;
+
+        jpeg_file = get_temporary_file ("simple-scan", "jpg");
+        result = page_save (page, "jpeg", jpeg_file, error);
+        jpeg_size = get_file_size (jpeg_file);
+        temporary_files = g_list_append (temporary_files, jpeg_file);
+
+        tiff_file = get_temporary_file ("simple-scan", "tiff");
+        result = page_save (page, "tiff", tiff_file, error);
+        tiff_size = get_file_size (tiff_file);
+        temporary_files = g_list_append (temporary_files, tiff_file);
+
+        /* Use the smallest file */
+        if (jpeg_size < tiff_size)
+            path = g_file_get_path (jpeg_file);
+        else
+            path = g_file_get_path (tiff_file);
+        g_string_append_printf (command_line, " %s", path);
+        g_free (path);
+    }
+
+    /* Use ImageMagick command to create a PDF */  
+    if (result) {
+        gchar *path, *stdout_text = NULL, *stderr_text = NULL;
+
+        output_file = get_temporary_file ("simple-scan", "pdf");
+        path = g_file_get_path (output_file);
+        g_string_append_printf (command_line, " %s", path);
+        g_free (path);
+
+        result = g_spawn_command_line_sync (command_line->str, &stdout_text, &stderr_text, &exit_status, error);
+        if (result && exit_status != 0) {
+            g_warning ("ImageMagick returned error code %d, command line was: %s", exit_status, command_line->str);
+            g_warning ("stdout: %s", stdout_text);
+            g_warning ("stderr: %s", stderr_text);
+            result = FALSE;
+        }
+        g_free (stdout_text);
+        g_free (stderr_text);
+    }
+
+    /* Move to target URI */
+    if (result) {
+        GFile *dest;
+        result = g_file_move (output_file, file, G_FILE_COPY_OVERWRITE, NULL, NULL, NULL, error);
+        g_object_unref (dest);
+    }
+  
+    /* Delete page files */
+    for (link = temporary_files; link; link = link->next) {
+        GFile *f = link->data;
+
+        g_file_delete (f, NULL, NULL);
+        g_object_unref (f);
+    }
+    g_list_free (temporary_files);
+
+    if (output_file)
+        g_object_unref (output_file);
+    g_string_free (command_line, TRUE);
+
+    return result;
+}
+
+
+static gboolean
+book_save_pdf (Book *book, GFile *file, GError **error)
 {
     GFileOutputStream *stream;
     GList *iter;
     cairo_surface_t *surface;
+    gchar *imagemagick_executable;
+  
+    /* Use ImageMagick if it is available as then we can compress the images */
+    imagemagick_executable = g_find_program_in_path ("convert");
+    if (imagemagick_executable) {
+        g_free (imagemagick_executable);
+        return book_save_pdf_with_imagemagick (book, file, error);
+    }
 
-    stream = open_file (uri, error);
+    stream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, error);
     if (!stream)
         return FALSE;
 
@@ -312,6 +396,24 @@ book_save_pdf (Book *book, const gchar *uri, GError **error)
 }
 
 
+gboolean
+book_save (Book *book, const gchar *type, GFile *file, GError **error)
+{
+    if (strcmp (type, "jpeg") == 0)
+        return book_save_multi_file (book, "jpeg", file, error);
+    else if (strcmp (type, "png") == 0)
+        return book_save_multi_file (book, "png", file, error);
+    else if (strcmp (type, "tiff") == 0)
+        return book_save_multi_file (book, "tiff", file, error);    
+    else if (strcmp (type, "ps") == 0)
+        return book_save_ps (book, file, error);    
+    else if (strcmp (type, "pdf") == 0)
+        return book_save_pdf (book, file, error);
+    else
+        return FALSE;
+}
+
+
 static void
 book_finalize (GObject *object)
 {
