@@ -9,7 +9,9 @@
  * license.
  */
 
+#include <stdlib.h>
 #include <string.h>
+#include <math.h>
 #include <sane/sane.h>
 #include <sane/saneopts.h>
 #include <glib/gi18n.h>
@@ -42,10 +44,11 @@ typedef struct
 typedef struct
 {
     gchar *device;
-    gint dpi;
+    gdouble dpi;
     ScanMode scan_mode;
     gint depth;
     gboolean type;
+    gint page_width, page_height;
 } ScanJob;
 
 typedef struct
@@ -91,7 +94,7 @@ struct ScannerPrivate
     /* Handle to SANE device */
     SANE_Handle handle;
     gchar *current_device;
-
+  
     SANE_Parameters parameters;
 
     /* Last option read */
@@ -99,7 +102,7 @@ struct ScannerPrivate
 
     /* Buffer for received line */
     SANE_Byte *buffer;
-    SANE_Int n_used;
+    SANE_Int buffer_size, n_used;
 
     SANE_Int bytes_remaining, line_count, pass_number, page_number, notified_page;
 
@@ -112,8 +115,6 @@ G_DEFINE_TYPE (Scanner, scanner, G_TYPE_OBJECT);
 /* Table of scanner objects for each thread (required for authorization callback) */
 static GHashTable *scanners;
 
-#define MAX_READ 10000
-
 
 static gboolean
 send_signal (SignalInfo *info)
@@ -196,7 +197,8 @@ set_scanning (Scanner *scanner, gboolean is_scanning)
 }
 
 
-static gint get_device_weight (const gchar *device)
+static gint
+get_device_weight (const gchar *device)
 {
     /* NOTE: This is using trends in the naming of SANE devices, SANE should be able to provide this information better */
   
@@ -385,36 +387,60 @@ static gboolean
 control_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int index, SANE_Action action, void *value)
 {
     SANE_Status status;
-    
+    gchar *old_value;
+
+    switch (option->type) {
+    case SANE_TYPE_BOOL:
+        old_value = g_strdup_printf (*((SANE_Bool *) value) ? "SANE_TRUE" : "SANE_FALSE");
+        break;
+    case SANE_TYPE_INT:
+        old_value = g_strdup_printf ("%d", *((SANE_Int *) value));
+        break;
+    case SANE_TYPE_FIXED:
+        old_value = g_strdup_printf ("%f", SANE_UNFIX (*((SANE_Fixed *) value)));
+        break;
+    case SANE_TYPE_STRING:
+        old_value = g_strdup_printf ("\"%s\"", (char *) value);
+        break;
+    default:
+        old_value = g_strdup ("?");
+        break;
+    }
+
     status = sane_control_option (handle, index, action, value, NULL);
     switch (option->type) {
     case SANE_TYPE_BOOL:
-        g_debug ("sane_control_option (%d, %s, %s) -> %s",
+        g_debug ("sane_control_option (%d, %s, %s) -> (%s, %s)",
                  index, get_action_string (action),
                  *((SANE_Bool *) value) ? "SANE_TRUE" : "SANE_FALSE",
-                 get_status_string (status));
+                 get_status_string (status),
+                 old_value);
         break;
     case SANE_TYPE_INT:
-        g_debug ("sane_control_option (%d, %s, %d) -> %s",
+        g_debug ("sane_control_option (%d, %s, %d) -> (%s, %s)",
                  index, get_action_string (action),
                  *((SANE_Int *) value),
-                 get_status_string (status));
+                 get_status_string (status),
+                 old_value);
         break;
     case SANE_TYPE_FIXED:
-        g_debug ("sane_control_option (%d, %s, %f) -> %s",
+        g_debug ("sane_control_option (%d, %s, %f) -> (%s, %s)",
                  index, get_action_string (action),
                  SANE_UNFIX (*((SANE_Fixed *) value)),
-                 get_status_string (status));
+                 get_status_string (status),
+                 old_value);
         break;
     case SANE_TYPE_STRING:
-        g_debug ("sane_control_option (%d, %s, \"%s\") -> %s",
+        g_debug ("sane_control_option (%d, %s, \"%s\") -> (%s, %s)",
                  index, get_action_string (action),
                  (char *) value,
-                 get_status_string (status));
+                 get_status_string (status),
+                 old_value);
         break;
     default:
         break;
     }
+    g_free (old_value);
 
     if (status != SANE_STATUS_GOOD)
         g_warning ("Error setting option %s: %s", option->name, sane_strstatus(status));
@@ -439,16 +465,18 @@ set_default_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SA
 
 
 static void
-set_bool_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Bool value)
+set_bool_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Bool value, SANE_Bool *result)
 {
     SANE_Bool v = value;
     g_return_if_fail (option->type == SANE_TYPE_BOOL);
     control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
+    if (result)
+        *result = v;
 }
 
 
 static void
-set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Int value)
+set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Int value, SANE_Int *result)
 {
     SANE_Int v = value;
 
@@ -464,32 +492,30 @@ set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_I
     }
     else if (option->constraint_type == SANE_CONSTRAINT_WORD_LIST) {
         int i;
-        SANE_Int min = INT_MIN, max = INT_MAX;
-      
-        /* Find nearest value above and below requested */
+        SANE_Int distance = INT_MAX, nearest = 0;
+
+        /* Find nearest value to requested */
         for (i = 0; i < option->constraint.word_list[0]; i++) {
             SANE_Int x = option->constraint.word_list[i+1];
-            if (x <= v && x > min)
-                min = x;
-            if (x >= v && x < max)
-                max = x;
+            if (abs (x - v) < distance) {
+                distance = abs (x - v);
+                nearest = x;
+            }
         }
-      
-        /* Pick nearest */
-        if (max - v < v - min)
-            v = max;
-        else
-            v = min;
+        v = nearest;
     }
 
     control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
+    if (result)
+        *result = v;
 }
 
 
 static void
-set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, double value)
+set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, gdouble value, gdouble *result)
 {
-    SANE_Fixed v;
+    gdouble v = value;
+    SANE_Fixed v_fixed;
 
     g_return_if_fail (option->type == SANE_TYPE_FIXED);
 
@@ -497,42 +523,39 @@ set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE
         double min = SANE_UNFIX (option->constraint.range->min);
         double max = SANE_UNFIX (option->constraint.range->max);
 
-        if (value < min)
-            value = min;
-        if (value > max)
-            value = max;
+        if (v < min)
+            v = min;
+        if (v > max)
+            v = max;
     }
     else if (option->constraint_type == SANE_CONSTRAINT_WORD_LIST) {
         int i;
-        double min = DBL_MIN, max = DBL_MAX;
-      
-        /* Find nearest value above and below requested */
+        double distance = DBL_MAX, nearest = 0.0;
+
+        /* Find nearest value to requested */
         for (i = 0; i < option->constraint.word_list[0]; i++) {
             double x = SANE_UNFIX (option->constraint.word_list[i+1]);
-            if (x <= value && x > min)
-                min = x;
-            if (x >= value && x < max)
-                max = x;
+            if (fabs (x - v) < distance) {
+                distance = fabs (x - v);
+                nearest = x;
+            }
         }
-      
-        /* Pick nearest */
-        if (max - value < value - min)
-            value = max;
-        else
-            value = min;
+        v = nearest;
     }
 
-    v = SANE_FIX (value);
-    control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
+    v_fixed = SANE_FIX (v);
+    control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v_fixed);
+    if (result)
+        *result = SANE_UNFIX (v_fixed);
 }
 
 
 static gboolean
-set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, const char *value)
+set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, const char *value, char **result)
 {
     char *string;
     gsize value_size, size;
-    gboolean result;
+    gboolean error;
 
     g_return_val_if_fail (option->type == SANE_TYPE_STRING, FALSE);
     
@@ -540,15 +563,18 @@ set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SAN
     size = option->size > value_size ? option->size : value_size;
     string = g_malloc(sizeof(char) * size);
     strcpy (string, value);
-    result = control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, string);
-    g_free (string);
+    error = control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, string);
+    if (result)
+        *result = string;
+    else
+        g_free (string);
    
-    return result;
+    return error;
 }
 
 
 static gboolean
-set_constrained_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, const char *values[])
+set_constrained_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, const char *values[], char **result)
 {
     gint i, j;
 
@@ -562,7 +588,7 @@ set_constrained_string_option (SANE_Handle handle, const SANE_Option_Descriptor
         }
 
         if (option->constraint.string_list[j] != NULL)
-            return set_string_option (handle, option, option_index, values[i]);
+            return set_string_option (handle, option, option_index, values[i], result);
     }
   
     return FALSE;
@@ -922,12 +948,12 @@ do_get_option (Scanner *scanner)
 
     if (strcmp (option->name, SANE_NAME_SCAN_RESOLUTION) == 0) {
         if (option->type == SANE_TYPE_FIXED) {
-            double dpi = job->dpi;
-            set_fixed_option (scanner->priv->handle, option, option_index, dpi);
+            set_fixed_option (scanner->priv->handle, option, option_index, job->dpi, &job->dpi);
         }
         else {
-            SANE_Int dpi = job->dpi;
-            set_int_option (scanner->priv->handle, option, option_index, dpi);
+            SANE_Int dpi;
+            set_int_option (scanner->priv->handle, option, option_index, job->dpi, &dpi);
+            job->dpi = dpi;
         }
     }
     else if (strcmp (option->name, SANE_NAME_SCAN_SOURCE) == 0) {
@@ -975,29 +1001,29 @@ do_get_option (Scanner *scanner)
         switch (job->type) {
         case SCAN_SINGLE:
             if (!set_default_option (scanner->priv->handle, option, option_index))
-                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, flatbed_sources))
+                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, flatbed_sources, NULL))
                     g_warning ("Unable to set single page source, please file a bug");
             break;
         case SCAN_ADF_FRONT:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_front_sources))
-                if (!!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources))                
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_front_sources, NULL))
+                if (!!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources, NULL))                
                     g_warning ("Unable to set front ADF source, please file a bug");
             break;
         case SCAN_ADF_BACK:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_back_sources))
-                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources))
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_back_sources, NULL))
+                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources, NULL))
                     g_warning ("Unable to set back ADF source, please file a bug");
             break;
         case SCAN_ADF_BOTH:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_duplex_sources))
-                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources))
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_duplex_sources, NULL))
+                if (!set_constrained_string_option (scanner->priv->handle, option, option_index, adf_sources, NULL))
                     g_warning ("Unable to set duplex ADF source, please file a bug");
             break;
         }
     }
     else if (strcmp (option->name, SANE_NAME_BIT_DEPTH) == 0) {
         if (job->depth > 0)
-            set_int_option (scanner->priv->handle, option, option_index, job->depth);
+            set_int_option (scanner->priv->handle, option, option_index, job->depth, NULL);
     }
     else if (strcmp (option->name, SANE_NAME_SCAN_MODE) == 0) {
         /* The names of scan modes often used in drivers, as taken from the sane-backends source */
@@ -1036,15 +1062,15 @@ do_get_option (Scanner *scanner)
             
         switch (job->scan_mode) {
         case SCAN_MODE_COLOR:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, color_scan_modes))
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, color_scan_modes, NULL))
                 g_warning ("Unable to set Color mode, please file a bug");
             break;
         case SCAN_MODE_GRAY:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, gray_scan_modes))
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, gray_scan_modes, NULL))
                 g_warning ("Unable to set Gray mode, please file a bug");
             break;
         case SCAN_MODE_LINEART:
-            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, lineart_scan_modes))
+            if (!set_constrained_string_option (scanner->priv->handle, option, option_index, lineart_scan_modes, NULL))
                 g_warning ("Unable to set Lineart mode, please file a bug");
             break;
         default:
@@ -1062,7 +1088,7 @@ do_get_option (Scanner *scanner)
             NULL
         };
             
-        if (!set_constrained_string_option (scanner->priv->handle, option, option_index, disable_compression_names))
+        if (!set_constrained_string_option (scanner->priv->handle, option, option_index, disable_compression_names, NULL))
             g_warning ("Unable to disable compression, please file a bug");
     }
     /* Always use maximum scan area - some scanners default to using partial areas.  This should be patched in sane-backends */
@@ -1072,31 +1098,47 @@ do_get_option (Scanner *scanner)
         {
         case SANE_CONSTRAINT_RANGE:
             if (option->type == SANE_TYPE_FIXED)
-                set_fixed_option (scanner->priv->handle, option, option_index, SANE_UNFIX (option->constraint.range->max));
+                set_fixed_option (scanner->priv->handle, option, option_index, SANE_UNFIX (option->constraint.range->max), NULL);
             else
-                set_int_option (scanner->priv->handle, option, option_index, option->constraint.range->max);
+                set_int_option (scanner->priv->handle, option, option_index, option->constraint.range->max, NULL);
             break;
         default:
             break;
         }
     }
+    else if (strcmp (option->name, SANE_NAME_PAGE_WIDTH) == 0) {
+        if (job->page_width > 0.0) {
+            if (option->type == SANE_TYPE_FIXED)
+                set_fixed_option (scanner->priv->handle, option, option_index, job->page_width / 10.0, NULL);
+            else
+                set_int_option (scanner->priv->handle, option, option_index, job->page_width / 10, NULL);
+        }
+    }
+    else if (strcmp (option->name, SANE_NAME_PAGE_HEIGHT) == 0) {
+        if (job->page_height > 0.0) {
+            if (option->type == SANE_TYPE_FIXED)
+                set_fixed_option (scanner->priv->handle, option, option_index, job->page_height / 10.0, NULL);
+            else 
+                set_int_option (scanner->priv->handle, option, option_index, job->page_height / 10, NULL);
+        }
+    }
 
     /* Test scanner options (hoping will not effect other scanners...) */
     if (strcmp (scanner->priv->current_device, "test") == 0) {
         if (strcmp (option->name, "hand-scanner") == 0) {
-            set_bool_option (scanner->priv->handle, option, option_index, FALSE);
+            set_bool_option (scanner->priv->handle, option, option_index, FALSE, NULL);
         }
         else if (strcmp (option->name, "three-pass") == 0) {
-            set_bool_option (scanner->priv->handle, option, option_index, FALSE);
+            set_bool_option (scanner->priv->handle, option, option_index, FALSE, NULL);
         }                    
         else if (strcmp (option->name, "test-picture") == 0) {
-            set_string_option (scanner->priv->handle, option, option_index, "Color pattern");
+            set_string_option (scanner->priv->handle, option, option_index, "Color pattern", NULL);
         }
         else if (strcmp (option->name, "read-delay") == 0) {
-            set_bool_option (scanner->priv->handle, option, option_index, TRUE);
+            set_bool_option (scanner->priv->handle, option, option_index, TRUE, NULL);
         }
         else if (strcmp (option->name, "read-delay-duration") == 0) {
-            set_int_option (scanner->priv->handle, option, option_index, 200000);
+            set_int_option (scanner->priv->handle, option, option_index, 200000, NULL);
         }
     }
 }
@@ -1183,7 +1225,8 @@ do_get_parameters (Scanner *scanner)
     info->width = scanner->priv->parameters.pixels_per_line;
     info->height = scanner->priv->parameters.lines;
     info->depth = scanner->priv->parameters.depth;
-    info->dpi = job->dpi;
+    info->dpi = job->dpi; // FIXME: This is the requested DPI, not the actual DPI
+    info->device = g_strdup (scanner->priv->current_device);
 
     if (scanner->priv->page_number != scanner->priv->notified_page) {
         emit_signal (scanner, GOT_PAGE_INFO, info);
@@ -1191,7 +1234,8 @@ do_get_parameters (Scanner *scanner)
     }
 
     /* Prepare for read */
-    scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * MAX_READ);
+    scanner->priv->buffer_size = scanner->priv->parameters.bytes_per_line + 1; /* Use +1 so buffer is not resized if driver returns one line per read */
+    scanner->priv->buffer = g_malloc (sizeof(SANE_Byte) * scanner->priv->buffer_size);
     scanner->priv->n_used = 0;
     scanner->priv->line_count = 0;
     scanner->priv->pass_number = 0;
@@ -1208,9 +1252,6 @@ do_complete_page (Scanner *scanner)
   
     job = (ScanJob *) scanner->priv->job_queue->data;
 
-    sane_cancel (scanner->priv->handle);
-    g_debug ("sane_cancel ()");
-
     /* If multi-pass then scan another page */
     if (!scanner->priv->parameters.last_frame) {
         scanner->priv->pass_number++;
@@ -1227,6 +1268,9 @@ do_complete_page (Scanner *scanner)
         return;
     }
 
+    sane_cancel (scanner->priv->handle);
+    g_debug ("sane_cancel ()");
+
     do_complete_document (scanner);
 }
 
@@ -1236,9 +1280,10 @@ do_read (Scanner *scanner)
 {
     SANE_Status status;
     SANE_Int n_to_read, n_read;
+    gboolean full_read = FALSE;
 
     /* Read as many bytes as we expect */
-    n_to_read = MAX_READ - scanner->priv->n_used;
+    n_to_read = scanner->priv->buffer_size - scanner->priv->n_used;
 
     status = sane_read (scanner->priv->handle,
                         scanner->priv->buffer + scanner->priv->n_used,
@@ -1261,7 +1306,9 @@ do_read (Scanner *scanner)
                    _("Error communicating with scanner"));
         return;
     }
- 
+
+    if (scanner->priv->n_used == 0 && n_read == scanner->priv->buffer_size)
+        full_read = TRUE;
     scanner->priv->n_used += n_read;
   
     /* Feed out lines */
@@ -1304,7 +1351,11 @@ do_read (Scanner *scanner)
         else {
             int i, n_remaining;
 
-            scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * MAX_READ);
+            /* Increase buffer size if did full read */
+            if (full_read)
+                scanner->priv->buffer_size += scanner->priv->parameters.bytes_per_line;
+
+            scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * scanner->priv->buffer_size);
             n_remaining = scanner->priv->n_used - (line->n_lines * line->data_length);
             scanner->priv->n_used = 0;
             for (i = 0; i < n_remaining; i++) {
@@ -1418,13 +1469,12 @@ scanner_is_scanning (Scanner *scanner)
 
 
 void
-scanner_scan (Scanner *scanner, const char *device,
-              gint dpi, ScanMode scan_mode, gint depth, ScanType type)
+scanner_scan (Scanner *scanner, const char *device, ScanOptions *options)
 {
     ScanRequest *request;
     const gchar *type_string;
   
-    switch (type) {
+    switch (options->type) {
     case SCAN_SINGLE:
         type_string = "SCAN_SINGLE";
         break;
@@ -1442,15 +1492,17 @@ scanner_scan (Scanner *scanner, const char *device,
         return;
     }
 
-    g_debug ("scanner_scan (\"%s\", %d, %s)", device ? device : "(null)", dpi, type_string);
+    g_debug ("scanner_scan (\"%s\", %d, %s)", device ? device : "(null)", options->dpi, type_string);
     request = g_malloc0 (sizeof (ScanRequest));
     request->type = REQUEST_START_SCAN;
     request->job = g_malloc0 (sizeof (ScanJob));
     request->job->device = g_strdup (device);
-    request->job->dpi = dpi;
-    request->job->scan_mode = scan_mode;
-    request->job->depth = depth;
-    request->job->type = type;
+    request->job->dpi = options->dpi;
+    request->job->scan_mode = options->scan_mode;
+    request->job->depth = options->depth;
+    request->job->type = options->type;
+    request->job->page_width = options->paper_width;
+    request->job->page_height = options->paper_height;
     g_async_queue_push (scanner->priv->scan_queue, request);
 }
 
